/* block-aio.c
 *
 * libaio-based raw disk implementation.
 *
 * (c) 2006 Andrew Warfield and Julian Chesterfield
 *
 * NB: This code is not thread-safe.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation; or, when distributed
 * separately from the Linux kernel or incorporated into other
 * software packages, subject to the following license:
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this source file (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */


#include <errno.h>
#include <libaio.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include "tapdisk.h"

#define MAX_AIO_REQS (MAX_REQUESTS * MAX_SEGMENTS_PER_REQ)

struct pending_aio {
	td_callback_t cb;
	int id;
	void *private;
};

struct tdaio_state {
	int fd;
	
	/* libaio state */
	io_context_t       aio_ctx;
	struct iocb        iocb_list  [MAX_AIO_REQS];
	struct iocb       *iocb_free  [MAX_AIO_REQS];
	struct pending_aio pending_aio[MAX_AIO_REQS];
	int                iocb_free_count;
	struct iocb       *iocb_queue[MAX_AIO_REQS];
	int                iocb_queued;
	struct io_event    aio_events[MAX_AIO_REQS];

	pthread_t	   aio_thread;
	/* pipe fds for communication with the aio completion thread */
	int		   command_fd[2];
	int		   completion_fd[2];
};

#define IOCB_IDX(_s, _io) ((_io) - (_s)->iocb_list)

static void *tdaio_completion_thread(void *);

/*Get Image size, secsize*/
static int get_image_info(struct td_state *s, int fd)
{
	int ret;
	long size;
	unsigned long total_size;
	struct stat stat;

	ret = fstat(fd, &stat);
	if (ret != 0) {
		DPRINTF("ERROR: fstat failed, Couldn't stat image");
		return -EINVAL;
	}

	if (S_ISBLK(stat.st_mode)) {
		/*Accessing block device directly*/
		s->size = 0;
		if (ioctl(fd,BLKGETSIZE,&s->size)!=0) {
			DPRINTF("ERR: BLKGETSIZE failed, couldn't stat image");
			return -EINVAL;
		}

		DPRINTF("Image size: \n\tpre sector_shift  [%llu]\n\tpost "
			"sector_shift [%llu]\n",
			(long long unsigned)(s->size << SECTOR_SHIFT),
			(long long unsigned)s->size);

		/*Get the sector size*/
#if defined(BLKSSZGET)
		{
			s->sector_size = DEFAULT_SECTOR_SIZE;
			ioctl(fd, BLKSSZGET, &s->sector_size);
			
			if (s->sector_size != DEFAULT_SECTOR_SIZE)
				DPRINTF("Note: sector size is %ld (not %d)\n",
					s->sector_size, DEFAULT_SECTOR_SIZE);
		}
#else
		s->sector_size = DEFAULT_SECTOR_SIZE;
#endif

	} else {
		/*Local file? try fstat instead*/
		s->size = (stat.st_size >> SECTOR_SHIFT);
		s->sector_size = DEFAULT_SECTOR_SIZE;
		DPRINTF("Image size: \n\tpre sector_shift  [%llu]\n\tpost "
			"sector_shift [%llu]\n",
			(long long unsigned)(s->size << SECTOR_SHIFT),
			(long long unsigned)s->size);
	}

	if (s->size == 0) {		
		s->size =((uint64_t) 16836057);
		s->sector_size = DEFAULT_SECTOR_SIZE;
	}
	s->info = 0;

	return 0;
}

/* Open the disk file and initialize aio state. */
int tdaio_open (struct td_state *s, const char *name)
{
	int i, fd, ret = 0;
	struct tdaio_state *prv = (struct tdaio_state *)s->private;
	s->private = prv;

	DPRINTF("block-aio open('%s')", name);
	/* Initialize AIO */
	prv->iocb_free_count = MAX_AIO_REQS;
	prv->iocb_queued     = 0;
	
	prv->aio_ctx = (io_context_t) 0;
	ret = io_setup(MAX_AIO_REQS, &prv->aio_ctx);

	if (ret < 0) {
                if (ret == -EAGAIN) {
                        DPRINTF("Couldn't setup AIO context.  If you are "
                                "trying to concurrently use a large number "
                                "of blktap-based disks, you may need to "
                                "increase the system-wide aio request limit. "
                                "(e.g. 'echo echo 1048576 > /proc/sys/fs/"
                                "aio-max-nr')\n");
                } else {
                        DPRINTF("Couldn't setup AIO context.\n");
                }
		goto done;
	}

	for (i=0;i<MAX_AIO_REQS;i++)
		prv->iocb_free[i] = &prv->iocb_list[i];

	/* Open the file */
        fd = open(name, O_RDWR | O_DIRECT | O_LARGEFILE);

        if ( (fd == -1) && (errno == EINVAL) ) {

                /* Maybe O_DIRECT isn't supported. */
                fd = open(name, O_RDWR | O_LARGEFILE);
                if (fd != -1) DPRINTF("WARNING: Accessing image without"
                                     "O_DIRECT! (%s)\n", name);

        } else if (fd != -1) DPRINTF("open(%s) with O_DIRECT\n", name);
	
        if (fd == -1) {
		DPRINTF("Unable to open [%s] (%d)!\n", name, 0 - errno);
        	ret = 0 - errno;
        	goto done;
        }

        prv->fd = fd;

	pipe(prv->command_fd);
	pipe(prv->completion_fd);
	
	ret = pthread_create(&prv->aio_thread, NULL, 
			     tdaio_completion_thread, prv);
	ret = 0;
	write(prv->command_fd[1], &ret, sizeof(ret));

			     
	ret = get_image_info(s, fd);
done:
	return ret;	
}

int tdaio_queue_read(struct td_state *s, uint64_t sector,
			       int nb_sectors, char *buf, td_callback_t cb,
			       int id, void *private)
{
	struct   iocb *io;
	struct   pending_aio *pio;
	struct   tdaio_state *prv = (struct tdaio_state *)s->private;
	int      size    = nb_sectors * s->sector_size;
	uint64_t offset  = sector * (uint64_t)s->sector_size;
	long     ioidx;
	
	if (prv->iocb_free_count == 0)
		return -ENOMEM;
	io = prv->iocb_free[--prv->iocb_free_count];
	
	ioidx = IOCB_IDX(prv, io);
	pio = &prv->pending_aio[ioidx];
	pio->cb = cb;
	pio->id = id;
	pio->private = private;
	
	io_prep_pread(io, prv->fd, buf, size, offset);
	io->data = (void *)ioidx;
	
	prv->iocb_queue[prv->iocb_queued++] = io;
	
	return 0;
}
			
int tdaio_queue_write(struct td_state *s, uint64_t sector,
			       int nb_sectors, char *buf, td_callback_t cb,
			       int id, void *private)
{
	struct   iocb *io;
	struct   pending_aio *pio;
	struct   tdaio_state *prv = (struct tdaio_state *)s->private;
	int      size    = nb_sectors * s->sector_size;
	uint64_t offset  = sector * (uint64_t)s->sector_size;
	long     ioidx;
	
	if (prv->iocb_free_count == 0)
		return -ENOMEM;
	io = prv->iocb_free[--prv->iocb_free_count];
	
	ioidx = IOCB_IDX(prv, io);
	pio = &prv->pending_aio[ioidx];
	pio->cb = cb;
	pio->id = id;
	pio->private = private;
	
	io_prep_pwrite(io, prv->fd, buf, size, offset);
	io->data = (void *)ioidx;
	
	prv->iocb_queue[prv->iocb_queued++] = io;
	
	return 0;
}
			
int tdaio_submit(struct td_state *s)
{
	int ret;
	struct   tdaio_state *prv = (struct tdaio_state *)s->private;

	ret = io_submit(prv->aio_ctx, prv->iocb_queued, prv->iocb_queue);
	
	/* XXX: TODO: Handle error conditions here. */
	
	/* Success case: */
	prv->iocb_queued = 0;
	
	return ret;
}

int *tdaio_get_fd(struct td_state *s)
{
	struct tdaio_state *prv = (struct tdaio_state *)s->private;
	int *fds, i;

	fds = malloc(sizeof(int) * MAX_IOFD);
	/*initialise the FD array*/
	for(i=0;i<MAX_IOFD;i++) fds[i] = 0;

	fds[0] = prv->completion_fd[0];

	return fds;	
}

int tdaio_close(struct td_state *s)
{
	struct tdaio_state *prv = (struct tdaio_state *)s->private;
	
	io_destroy(prv->aio_ctx);
	close(prv->fd);
	
	return 0;
}

/* 
 * We don't have any way to do epoll on aio events in a normal kernel, so
 * wait for aio events in a separate thread and return completion status 
 * that via a pipe that can be waited on normally.
 * 
 * To keep locking problems between the completion thread and the submit
 * thread to a minimum, there's a handshake which allows only one thread
 * to be doing work on the completion queue at a time:
 * 
 * 1) main thread sends completion thread a command via the command pipe;
 * 2) completion thread waits for aio events and returns the number 
 *    received on the completion pipe
 * 3) main thread processes the received prv->aio_events events
 * 4) loop back to 1) to let the completion thread refill the aio_events 
 *    buffer.
 *
 * This workaround needs to disappear once the kernel provides a single
 * mechanism for waiting on both aio and normal fd wakeups.
 */

static void *tdaio_completion_thread(void *arg)
{
	struct tdaio_state *prv = (struct tdaio_state *) arg;
	int command;
	int nr_events;
	int rc;
	
	while (1) {
		rc = read(prv->command_fd[0], &command, sizeof(command));

		do {
			/* Non-blocking test for completed io. */
			rc = io_getevents(prv->aio_ctx, 1, 
					  MAX_AIO_REQS, prv->aio_events,
					  NULL);
			if (rc) {
				nr_events = rc;
				rc = write(prv->completion_fd[1], &nr_events, 
					   sizeof(nr_events));
			}
		} while (!rc);
	}
}

int tdaio_do_callbacks(struct td_state *s, int sid)
{
	int ret, i, nr_events, rsp = 0;
	struct io_event *ep;
	struct tdaio_state *prv = (struct tdaio_state *)s->private;

	/* Non-blocking test for completed io. */
	ret = read(prv->completion_fd[0], &nr_events, sizeof(nr_events));

repeat:	
	for (ep=prv->aio_events,i=nr_events; i-->0; ep++) {
		struct iocb        *io  = ep->obj;
		struct pending_aio *pio;
		
		pio = &prv->pending_aio[(long)io->data];
		rsp += pio->cb(s, ep->res == io->u.c.nbytes ? 0 : 1,
			       pio->id, pio->private);

		prv->iocb_free[prv->iocb_free_count++] = io;
	}

	if (nr_events) {
		nr_events = io_getevents(prv->aio_ctx, 0, 
					 MAX_AIO_REQS, prv->aio_events,
					 NULL);
		goto repeat;
	}

	write(prv->command_fd[1], &nr_events, sizeof(nr_events));

	return rsp;
}
	
struct tap_disk tapdisk_aio = {
	"tapdisk_aio",
	sizeof(struct tdaio_state),
	tdaio_open,
	tdaio_queue_read,
	tdaio_queue_write,
	tdaio_submit,
	tdaio_get_fd,
	tdaio_close,
	tdaio_do_callbacks,
};
